# Generated by Django 4.2.3 on 2023-09-03 20:26

from django.db import migrations
import djangoldp.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0016_alter_activity_options_alter_follower_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='follower',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='ldpsource',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
    ]
